@Override
	public CombiningAlg.Evaluator getInstance(final Iterable<CombiningAlgParameter<? extends T>> params, final Iterable<? extends T> combinedElements)
	        throws UnsupportedOperationException, IllegalArgumentException
	{
		// if no element -> NotApplicable
		if (combinedElements == null)
		{
			LOGGER.warn("{}: no element to combine -> optimization: replacing with equivalent evaluator returning constant NotApplicable decision", this);
			return CombiningAlgEvaluators.NOT_APPLICABLE_CONSTANT_EVALUATOR;
		}

		final Iterator<? extends Decidable> combinedEltIterator = combinedElements.iterator();
		if (!combinedEltIterator.hasNext())
		{
			// empty (no element to combine)
			LOGGER.warn("{}: no element to combine -> optimization: replacing with equivalent evaluator returning constant NotApplicable decision", this);
			return CombiningAlgEvaluators.NOT_APPLICABLE_CONSTANT_EVALUATOR;
		}

		/*
		 * If combined elements are Rules, we can optimize
		 */
		if (!RuleEvaluator.class.isAssignableFrom(getCombinedElementType()))
		{
			return new OrderPreservingCombiningAlgEvaluator(combinedElements, this.overridingEffect);
		}

		// combined elements are Rules, we can optimize
		/*
		 * There is at least one Rule. Prepare to iterate over Rules and collect them in a specific way (depending on whether it is "ordered-*-overrides" kind of algorithm or not).
		 */
		final RuleCollector ruleCollector = ruleCollectorFactory.newInstance();

		/*
		 * 
		 * If we find any empty Rule in overridden Effect (no target/condition/pep_action), we don't need to look at other rules with such Effect and no PEP action; because if there is no rule with
		 * overriding Effect, this rule is enough to return the overridden Effect as decision, and PEP actions come from all other rules with same Effect and PEP actions (e.g. if algorithm is
		 * deny-overrides, then if there is no applicable Deny rule, if there is any empty Permit rule, the result is Permit with PEP actions combined from all other Permit rules with PEP actions)
		 */
		RuleEvaluator firstEmptyRuleWithOverriddenEffect = null;

		while (combinedEltIterator.hasNext())
		{
			final RuleEvaluator rule = (RuleEvaluator) combinedEltIterator.next();
			if (rule.getEffect() == overridingEffect)
			{
				/*
				 * If rule's effect is the overriding Effect and it has no target/condition/pep_actions, then rule will always return this Effect -> {overriding_effect}-overrides alg always evaluates
				 * to ${overriding_effect} (ignore/remove all other rules). ({overriding_effect} = Permit if algorithm is Permit-overrides, or Deny if algorithm is Deny-overrides in this statement.)
				 */
				if (rule.isEmptyEquivalent())
				{
					LOGGER.warn(
					        "{}: {} with Effect={} is empty (no target/condition/pep_actions) => always returns {} => algorithm will always return {} => other combined rules have no effect => will be ignored/removed.",
					        this, rule, this.overridingEffect, this.overridingEffect, this.overridingEffect);
					return constantOverridingEffectDecisionEvaluator;
				}

				/*
				 * Rule is not empty, i.e. has a target/condition/actions, therefore may not necessarily return its (overriding) Effect as decision
				 */
				ruleCollector.addNonEmptyRuleWithOverridingEffect(rule);
				continue;
			}

			/*
			 * Rule Effect = {overridden_Effect} (e.g. Permit if algorithm is deny-overrides)
			 * 
			 * In the end, if there is no applicable rule with overridING Effect, and if there is an empty rule with such overridden Effect, we already know that the result is always the overridden
			 * Effect with PEP actions from all other rules with same Effect and PEP action(s).
			 */
			if (firstEmptyRuleWithOverriddenEffect != null)
			{
				if (rule.hasAnyPepAction())
				{

					// rule has PEP action(s), therefore it matters if final result is overridden Effect
					ruleCollector.addRuleWithOverriddenEffectAndPepActions(rule);
					continue;
				}

				/*
				 * Rule has no PEP action Ignore this new Rule with overridden Effect and no PEP action; it will have no effect since we are sure the empty rule (that we already found) with overridden
				 * Effect will successfully evaluate.
				 */
				LOGGER.warn(
				        "{}: Ignoring/removing {} (Effect={}) because it does not affect the result, since it does no have any PEP action and we already found an empty Rule ({}) found with same Effect (always returns {}).",
				        this, rule, overriddenEffect, firstEmptyRuleWithOverriddenEffect, overriddenEffect);
				// continue looking for rules with overriding Effect
				continue;

			}

			// No empty Rule with overridden Effect found yet (firstEmptyRuleWithOverriddenEffect == null)
			if (rule.isEmptyEquivalent())
			{
				// rule has no PEP action -> firstEmptyRuleWithOverriddenEffect == null
				/*
				 * This is the first declared empty Rule with overridden Effect -> always returns the overridden Effect as decision; we can ignore/remove other Rules with overridden Effect unless they
				 * have PEP actions (have no effect anymore)
				 */
				LOGGER.warn(
				        "{}: {} (Effect={}) is empty (no target/condition/pep_actions) => always returns {} => algorithm will always return {} unless some {} rule applies => other combined {} rules without any PEP action have no effect => will be ignored/removed.",
				        this, rule, overriddenEffect, overriddenEffect, overriddenEffect, overridingEffect, overriddenEffect);
				firstEmptyRuleWithOverriddenEffect = rule;
				ruleCollector.addFirstEmptyRuleWithOverriddenEffect(rule);
				// continue looking for rules with overriding Effect or with PEP actions
				continue;
			}

			/*
			 * Non-empty Rule with overridden Effect found yet and current rule is not empty
			 */
			if (rule.hasAnyPepAction())
			{
				ruleCollector.addRuleWithOverriddenEffectAndPepActions(rule);
			} else
			{
				ruleCollector.addNonEmptyRuleWithOverriddenEffectButNoPepAction(rule);
			}

		} // END while

		/*
		 * There is at least one rule and there is no empty Rule with overriding Effect.
		 * 
		 * If there is no rule with overriding Effect...
		 */
		if (!ruleCollector.hasRuleWithOverridingEffect())
		{
			/*
			 * No Rule with overriding Effect (whether empty or not) -> at least one Rule with overridden Effect and all rules have this same overridden Effect. If we found an empty rule with
			 * overridden Effect and no other with PEP action, the decision is the constant overridden Effect without PEP action.
			 */
			if (firstEmptyRuleWithOverriddenEffect != null && !ruleCollector.hasRuleWithOverriddenEffectAndPepAction())
			{
				/*
				 * no Rule with overriding Effect or PEP action, but one empty rule with overridden Effect -> final result is the overridden Effect as simple decision (no PEP action) always
				 */
				LOGGER.warn(
				        "{}: the only combined rule is empty {} Rule ({}) => algorithm will always return this {} => optimization: replacing with equivalent evaluator returning constant {} decision",
				        this, this.overriddenEffect, firstEmptyRuleWithOverriddenEffect, this.overriddenEffect, this.overriddenEffect);
				return constantOverriddenEffectDecisionEvaluator;
			}

			/*
			 * (All rules have same overridden Effect, and) either there is no empty rule OR there is at least one with PEP action
			 */
			return ruleCollector.getRuleCombiningAlgEvaluatorAssumingAllWithOverriddenEffect();
		}

		/*
		 * There is at least one non-empty rule with overriding Effect.
		 * 
		 * If there is no rule with overridden Effect...
		 */

		if (!ruleCollector.hasRuleWithOverriddenEffect())
		{
			/*
			 * No rule with overridden Effect -> only non-empty rules with same overriding Effect
			 */
			return ruleCollector.getRuleCombiningAlgEvaluatorAssumingAllWithOverridingEffect();
		}

		/*
		 * At least one Rule with overridden Effect and only non-empty rules with overriding Effect
		 */
		return ruleCollector.getDPOverridesRuleCombiningAlgEvaluator(overridingEffect);
	}
