@Override
	public FirstOrderFunctionCall<BooleanValue> newCall(final List<Expression<?>> argExpressions, final Datatype<?>... remainingArgTypes) throws IllegalArgumentException {
		/*
		 * Optimization: check whether the first arg ('n' = number of Trues to reach) is constant
		 */
		final Iterator<? extends Expression<?>> argExpsIterator = argExpressions.iterator();
		/*
		 * Evaluate the first argument if not in remainingArgTypes
		 */
		if (!argExpsIterator.hasNext())
		{
			return new CallWithVarArg0(functionSignature, argExpressions, remainingArgTypes);
		}

		final Optional<? extends Value> arg0 = argExpsIterator.next().getValue();
		if (arg0.isPresent())
		{
			// arg0 is constant
			// We downsize the BigInteger value to int right away, because anyway inputs.size() is an
			// int, so we cannot do better and don't need to.
			final int nOfRequiredTrues = IntegerValue.class.cast(arg0.get()).getUnderlyingValue().intValueExact();
			if (nOfRequiredTrues < 0)
			{
				throw new IllegalArgumentException(getInvalidArg0MessagePrefix(functionSignature) + nOfRequiredTrues);
			}

			// If the number of trues needed is zero, return true.
			if (nOfRequiredTrues == 0)
			{
				return new ConstantResultFirstOrderFunctionCall<>(BooleanValue.TRUE, StandardDatatypes.BOOLEAN);
			}

			// else nOfRequiredTrues > 0
			// make sure it's possible to find n true values in the remaining arguments
			// Total number of args to the n-of function:
			final int totalArgCount = argExpressions.size() + remainingArgTypes.length;
			/*
			 * Number of remaining args, i.e. all (boolean) args after the first (integer) arg (excluded).
			 */
			final int nOfRemainingArgs = totalArgCount - 1;
			if (nOfRequiredTrues > nOfRemainingArgs)
			{
				throw new IllegalArgumentException(getInvalidArg0MessagePrefix(functionSignature) + nOfRequiredTrues + " > number_of_remaining args (" + nOfRemainingArgs + ")");
			}

			return new CallWithFixedArg0(functionSignature, nOfRequiredTrues, argExpressions, remainingArgTypes);
		}

		/*
		 * TODO: optimize by checking the following: If any argument expression is constant BooleanAttributeValue False, remove it from the arguments, as it has no effect on the final result. Indeed,
		 * n-of function is commutative except for the first argument, and n-of(N, false, x, y...) = n-of(N, x, y...).
		 */
		return new CallWithVarArg0(functionSignature, argExpressions, remainingArgTypes);
	}
