@Override
	public FirstOrderFunctionCall<AV> newCall(final List<Expression<?>> argExpressions, final Datatype<?>... remainingArgTypes) throws IllegalArgumentException {
		/*
		 * If this.op is a commutative function (e.g. add or multiply function), we can simplify arguments if there are multiple constants. Indeed, if C1,...Cm are constants, then:
		 * <p>
		 * op(x1,..., x_{n1-1}, C1, x_n1, ..., x_{n2-1}, C2, x_n2, ..., Cm, x_nm...) = op( C, x1.., x_{n1-1}, x_n1, x_{n2-2}, x_n2...), where C (constant) = op(C1, C2..., Cm)
		 * </p>
		 * In this case, we can pre-compute constant C and replace all constant args with one: C
		 * 
		 */
		if (op instanceof MultaryOperation && ((MultaryOperation<AV>) op).isCommutative())
		{
			/*
			 * Constant argExpressions
			 */
			final Deque<AV> constants = new ArrayDeque<>(argExpressions.size());
			/*
			 * Remaining variable argExpressions
			 */
			final List<Expression<?>> finalArgExpressions = new ArrayList<>(argExpressions.size());
			final Datatype<AV> paramType = this.functionSignature.getParameterType();
			final Iterator<Expression<?>> argExpIterator = argExpressions.iterator();
			int argIndex = 0;
			while (argExpIterator.hasNext())
			{
				final Expression<?> argExp = argExpIterator.next();
				final Optional<? extends Value> v = argExp.getValue();
				if (!v.isPresent())
				{
					// variable
					finalArgExpressions.add(argExp);
				} else
				{
					// constant
					try
					{
						constants.add(paramType.cast(v.get()));
					} catch (final ClassCastException e)
					{
						throw new IllegalArgumentException(
								"Function " + this.functionSignature + ": invalid arg #" + argIndex + ": bad type: " + argExp.getReturnType() + ". Expected type: " + paramType, e);
					}
				}

				argIndex += 1;
			}

			if (constants.size() > 1)
			{
				/*
				 * we can replace all constant args C1, C2... with one constant C = op(C1, C2...)
				 */
				LOGGER.warn("Function {}: simplifying args to this commutative function (f): replacing all constant args {} with one that is the constant result of f(constant_args)",
						this.functionSignature, constants);
				final AV constantResult = op.eval(constants);
				if (finalArgExpressions.isEmpty())
				{
					/*
					 * There aren't any other args, i.e. all are constant. The result is constantResult.
					 */
					return new ConstantResultFirstOrderFunctionCall<>(constantResult, paramType);
				}

				/*
				 * finalArgExpressions is not empty. There is at least one variable arg.
				 */
				finalArgExpressions.add(new ConstantPrimitiveAttributeValueExpression<>(paramType, constantResult));
				return new Call<>(functionSignature, op, finalArgExpressions, remainingArgTypes);
			}

		}

		return new Call<>(functionSignature, op, argExpressions, remainingArgTypes);
	}
