@Override
	public FirstOrderFunctionCall<StringValue> newCall(final List<Expression<?>> argExpressions, final Datatype<?>... remainingArgTypes)
	{
		final int numOfArgs = argExpressions.size() + remainingArgTypes.length;
		if (numOfArgs != 3)
		{
			throw new IllegalArgumentException("Function " + functionSignature.getName() + ": Invalid number of args: expected: 3; actual: " + numOfArgs);
		}

		/*
		 * Datatypes will be checked by Call class but they are specific constraints: arg0 >= 0 && (arg1 == -1 || arg0 <= arg1). Note that if arg0 == arg1, result may be '' or raise out-of-bounds
		 * error
		 */
		final Iterator<? extends Expression<?>> argExpsIterator = argExpressions.iterator();
		if (!argExpsIterator.hasNext())
		{
			return new Call(functionSignature, param0Type, argExpressions, remainingArgTypes);
		}

		// Skip the first argument (arg #0) which is the string
		argExpsIterator.next();

		// Second arg (arg #1)
		if (!argExpsIterator.hasNext())
		{
			return new Call(functionSignature, param0Type, argExpressions, remainingArgTypes);
		}

		final Expression<?> arg1Exp = argExpsIterator.next();
		final Optional<? extends Value> arg1 = arg1Exp.getValue();
		final int beginIndex;
		if (arg1.isPresent())
		{
			final Value arg1Value = arg1.get();
			if (!(arg1Value instanceof IntegerValue))
			{
				throw new IllegalArgumentException(getInvalidArg1MessagePrefix(functionSignature) + arg1Value + " (type: " + arg1Exp.getReturnType() + ")");
			}

			beginIndex = IntegerValue.class.cast(arg1Value).getUnderlyingValue().intValueExact();
			if (beginIndex < 0)
			{
				throw new IllegalArgumentException(getInvalidArg1MessagePrefix(functionSignature) + beginIndex);
			}
		} else
		{
			beginIndex = -1; // undefined
		}

		// Third arg
		if (!argExpsIterator.hasNext())
		{
			return new Call(functionSignature, param0Type, argExpressions, remainingArgTypes);
		}

		final Expression<?> arg2Exp = argExpsIterator.next();
		final Optional<? extends Value> arg2 = arg2Exp.getValue();
		if (arg2.isPresent())
		{
			final Value arg2Value = arg2.get();
			if (!(arg2Value instanceof IntegerValue))
			{
				throw new IllegalArgumentException(getInvalidArg2MessagePrefix(functionSignature) + arg2Value + " (type: " + arg2Exp.getReturnType() + ")");
			}

			final int endIndex = IntegerValue.class.cast(arg2Value).getUnderlyingValue().intValueExact();
			if (endIndex != -1)
			{
				if (endIndex < 0)
				{
					throw new IllegalArgumentException(getInvalidArg2MessagePrefix(functionSignature) + endIndex);
				}

				if (beginIndex != -1)
				{
					// beginIndex defined
					if (endIndex < beginIndex)
					{
						// and endIndex strictly smaller than beginIndex!
						throw new IllegalArgumentException(getInvalidArg2MessagePrefix(functionSignature) + endIndex + " < beginIndex (" + beginIndex + ")");
					}

					// endIndex >= beginIndex
					if (beginIndex == 0 && endIndex == 0)
					{
						// constant empty string
						LOGGER.warn("{} used with arg0 (beginIndex) = arg1 (endIndex) = 0 resulting in constant empty string. This is useless!", this.functionSignature);
						return new ConstantResultFirstOrderFunctionCall<>(StringValue.EMPTY, StandardDatatypes.STRING);
					}
				}

			}
		}

		return new Call(functionSignature, param0Type, argExpressions, remainingArgTypes);
	}
