public static StaticTopLevelPolicyElementEvaluator getInstance(final Policy policyElement, final XPathCompiler parentDefaultXPathCompiler, final Map<String, String> namespacePrefixesByURI,
	        final ExpressionFactory expressionFactory, final CombiningAlgRegistry combiningAlgRegistry) throws IllegalArgumentException
	{
		if (policyElement == null)
		{
			throw NULL_XACML_POLICY_ARG_EXCEPTION;
		}

		if (expressionFactory == null)
		{
			throw NULL_EXPRESSION_FACTORY_EXCEPTION;
		}

		if (combiningAlgRegistry == null)
		{
			throw NULL_XACML_COMBINING_ALG_ARG_EXCEPTION;
		}

		final String policyId = policyElement.getPolicyId();
		final PolicyVersion policyVersion = new PolicyVersion(policyElement.getVersion());
		final String policyFriendlyId = "Policy[" + policyId + "#v" + policyVersion + "]";
		final DefaultsType policyDefaults = policyElement.getPolicyDefaults();

		/*
		 * Inherited PolicyDefaults is this.policyDefaults if not null, the parentPolicyDefaults otherwise
		 */
		final XPathCompiler defaultXPathCompiler;
		if (policyDefaults == null)
		{
			defaultXPathCompiler = parentDefaultXPathCompiler;
		}
		else
		{
			try
			{
				defaultXPathCompiler = XmlUtils.newXPathCompiler(policyDefaults.getXPathVersion(), namespacePrefixesByURI);
			}
			catch (final IllegalArgumentException e)
			{
				throw new IllegalArgumentException(policyFriendlyId + ": Invalid PolicyDefaults/XPathVersion or XML namespace prefix/URI undefined", e);
			}

		}

		/*
		 * Elements defined in xs:choice of XACML schema type PolicyType: Rules/(Rule)CombinerParameters/VariableDefinitions
		 */
		final List<Serializable> policyChoiceElements = policyElement.getCombinerParametersAndRuleCombinerParametersAndVariableDefinitions();
		/*
		 * There are at most as many combining alg parameters as policyChoiceElements.size().
		 */
		final List<CombiningAlgParameter<? extends RuleEvaluator>> combiningAlgParameters = new ArrayList<>(policyChoiceElements.size());

		/*
		 * Keep a copy of locally-defined variable IDs defined in this policy, to remove them from the global manager at the end of parsing this policy. They should not be visible outside the scope of
		 * this policy. There are at most as many VariableDefinitions as policyChoiceElements.size().
		 */
		final List<String> localVariableIds = new ArrayList<>(policyChoiceElements.size());
		/*
		 * We keep a record of the size of the longest chain of VariableReference in this policy, and update it when a VariableDefinition occurs
		 */
		int sizeOfPolicyLongestVarRefChain = 0;
		/*
		 * Map to get rules by their ID so that we can resolve rules associated with RuleCombinerParameters, and detect duplicate RuleId. We want to preserve insertion order, to get map.values() in
		 * order of declaration, so that ordered-* algorithms have rules in order. There are at most as many Rules as policyChoiceElements.size().
		 */
		final Map<String, RuleEvaluator> ruleEvaluatorsByRuleIdInOrderOfDeclaration = new LinkedHashMap<>(policyChoiceElements.size());
		int childIndex = 0;
		for (final Serializable policyChildElt : policyChoiceElements)
		{
			if (policyChildElt instanceof RuleCombinerParameters)
			{
				final String combinedRuleId = ((RuleCombinerParameters) policyChildElt).getRuleIdRef();
				final RuleEvaluator ruleEvaluator = ruleEvaluatorsByRuleIdInOrderOfDeclaration.get(combinedRuleId);
				if (ruleEvaluator == null)
				{
					throw new IllegalArgumentException(
					        policyFriendlyId + ":  invalid RuleCombinerParameters: referencing undefined child Rule #" + combinedRuleId + " (no such rule defined before this element)");
				}

				final BaseCombiningAlgParameter<RuleEvaluator> combiningAlgParameter;
				try
				{
					combiningAlgParameter = new BaseCombiningAlgParameter<>(ruleEvaluator, ((CombinerParametersType) policyChildElt).getCombinerParameters(), expressionFactory, defaultXPathCompiler);
				}
				catch (final IllegalArgumentException e)
				{
					throw new IllegalArgumentException(policyFriendlyId + ": invalid child #" + childIndex + " (RuleCombinerParameters)", e);
				}

				combiningAlgParameters.add(combiningAlgParameter);
			}
			else if (policyChildElt instanceof CombinerParametersType)
			{
				/*
				 * CombinerParameters that is not RuleCombinerParameters already tested before
				 */
				final BaseCombiningAlgParameter<RuleEvaluator> combiningAlgParameter;
				try
				{
					combiningAlgParameter = new BaseCombiningAlgParameter<>(null, ((CombinerParametersType) policyChildElt).getCombinerParameters(), expressionFactory, defaultXPathCompiler);
				}
				catch (final IllegalArgumentException e)
				{
					throw new IllegalArgumentException(policyFriendlyId + ": invalid child #" + childIndex + " (CombinerParameters)", e);
				}

				combiningAlgParameters.add(combiningAlgParameter);
			}
			else if (policyChildElt instanceof VariableDefinition)
			{
				final VariableDefinition varDef = (VariableDefinition) policyChildElt;
				final Deque<String> varDefLongestVarRefChain = new ArrayDeque<>();
				final VariableReference<?> var;
				try
				{
					var = expressionFactory.addVariable(varDef, defaultXPathCompiler, varDefLongestVarRefChain);
				}
				catch (final IllegalArgumentException e)
				{
					throw new IllegalArgumentException(policyFriendlyId + ": invalid child #" + childIndex + " (VariableDefinition)", e);
				}

				if (var != null)
				{
					/*
					 * Conflicts can occur between variables defined in this policy but also with others already in a wider scope, i.e. defined in parent/ancestor policy
					 */
					throw new IllegalArgumentException(policyFriendlyId + ": Duplicable VariableDefinition for VariableId = " + var.getVariableId());
				}

				localVariableIds.add(varDef.getVariableId());
				/*
				 * check whether the longest VariableReference chain in the VariableDefinition is longer than what we've got so far
				 */
				final int sizeOfVarDefLongestVarRefChain = varDefLongestVarRefChain.size();
				if (sizeOfVarDefLongestVarRefChain > sizeOfPolicyLongestVarRefChain)
				{
					sizeOfPolicyLongestVarRefChain = sizeOfVarDefLongestVarRefChain;
				}
			}
			else if (policyChildElt instanceof Rule)
			{
				final RuleEvaluator ruleEvaluator;
				try
				{
					ruleEvaluator = new RuleEvaluator((Rule) policyChildElt, defaultXPathCompiler, expressionFactory);
				}
				catch (final IllegalArgumentException e)
				{
					throw new IllegalArgumentException(policyFriendlyId + ": Error parsing child #" + childIndex + " (Rule)", e);
				}

				final RuleEvaluator conflictingRuleEvaluator = ruleEvaluatorsByRuleIdInOrderOfDeclaration.putIfAbsent(ruleEvaluator.getRuleId(), ruleEvaluator);
				if (conflictingRuleEvaluator != null)
				{
					/*
					 * Conflict: 2 Rule elements with same RuleId -> violates uniqueness of RuleId within a Policy (XACML spec)
					 */
					throw new IllegalArgumentException(policyFriendlyId + ": Duplicate Rule with RuleId = " + conflictingRuleEvaluator.getRuleId());
				}
			}

			childIndex++;
		}

		final PrimaryPolicyMetadata primaryPolicyMetadata = new BasePrimaryPolicyMetadata(TopLevelPolicyElementType.POLICY, policyId, policyVersion);
		final ObligationExpressions obligationExps = policyElement.getObligationExpressions();
		final AdviceExpressions adviceExps = policyElement.getAdviceExpressions();
		final StaticTopLevelPolicyElementEvaluator policyEvaluator = new StaticBaseTopLevelPolicyElementEvaluator<>(RuleEvaluator.class, primaryPolicyMetadata, Optional.empty(),
		        policyElement.getTarget(), policyElement.getRuleCombiningAlgId(), ruleEvaluatorsByRuleIdInOrderOfDeclaration.values(), combiningAlgParameters,
		        obligationExps == null ? null : obligationExps.getObligationExpressions(), adviceExps == null ? null : adviceExps.getAdviceExpressions(),
		        Collections.<String>unmodifiableList(localVariableIds), defaultXPathCompiler, expressionFactory, combiningAlgRegistry);

		/*
		 * We are done parsing expressions in this policy, including VariableReferences, it's time to remove variables scoped to this policy from the variable manager
		 */
		for (final String varId : localVariableIds)
		{
			expressionFactory.removeVariable(varId);
		}

		return policyEvaluator;
	}
