private PolicyQueryResult getJaxbPolicyElement(final String policyTypeId, final String policyId, final Optional<PolicyVersionPatterns> policyPolicyVersionPatterns)
	        throws IndeterminateEvaluationException
	{
		final Optional<PolicyVersionPattern> versionPattern;
		if (policyPolicyVersionPatterns.isPresent())
		{
			/*
			 * TODO: the following code does not support LatestVersion and EarliestVersion patterns. Beware that comparing versions (XACML VersionType) to each other - and also comparing literal
			 * version to version pattern (XACML VersionMatchType) - is NOT the same as sorting strings in lexicographical order or matching standard regular expressions. Indeed, in XACML, a version
			 * (VersionType) is a sequence/array of decimal numbers actually, therefore it relies on number comparison; and version pattern use wildcard characters '*' and '+' with a special meaning
			 * that is different from PCRE or other regex engines.
			 */
			final PolicyVersionPatterns nonNullPolicyPolicyVersionPatterns = policyPolicyVersionPatterns.get();
			if (nonNullPolicyPolicyVersionPatterns.getEarliestVersionPattern().isPresent())
			{
				throw new IllegalArgumentException("PolicyProvider '" + id + "': EarliestVersion in input policy reference is not supported");
			}

			if (nonNullPolicyPolicyVersionPatterns.getLatestVersionPattern().isPresent())
			{
				throw new IllegalArgumentException("PolicyProvider '" + id + "': LatestVersion in input policy reference is not supported");
			}

			versionPattern = nonNullPolicyPolicyVersionPatterns.getVersionPattern();
		}
		else
		{
			versionPattern = Optional.empty();
		}

		final PolicyPojo policyPOJO;
		/*
		 * TODO: the following code will get any policy version that matches the policy type, id and optional VersionMatch. It may be smarter to always get the latest if there are multiple matches.
		 * But this adds complexity as mentioned in previous TODO comment.
		 */
		if (versionPattern.isPresent())
		{
			final PolicyVersionPattern nonNullVersionPattern = versionPattern.get();
			final PolicyVersion versionLiteral = nonNullVersionPattern.toLiteral();
			if (versionLiteral != null)
			{
				policyPOJO = policyCollection.findOne("{type: #, id: #, version: #}", policyTypeId, policyId, versionLiteral.toString()).as(PolicyPojo.class);
			}
			else
			{
				/*
				 * versionPattern is not a literal/constant version (contains wildcard '*' or '+') -> convert to PCRE regex for MongoDB server-side evaluation
				 */
				final String regex = "^" + nonNullVersionPattern.toRegex() + "$";
				policyPOJO = policyCollection.findOne("{type: #, id: #, version: { $regex: # }}", policyTypeId, policyId, regex).as(PolicyPojo.class);
			}
		}
		else
		{
			// no version pattern specified
			policyPOJO = policyCollection.findOne("{type: #, id: #}", policyTypeId, policyId).as(PolicyPojo.class);
		}

		if (policyPOJO == null)
		{
			return null;
		}

		final XmlnsFilteringParser xacmlParser;
		try
		{
			xacmlParser = xacmlParserFactory.getInstance();
		}
		catch (final JAXBException e)
		{
			throw new IndeterminateEvaluationException("PolicyProvider " + id + ": Failed to create JAXB unmarshaller for XACML Policy(Set)", XacmlStatusCode.PROCESSING_ERROR.value(), e);
		}

		final InputSource xmlInputSrc = new InputSource(new StringReader(policyPOJO.getContent()));
		final Object resultJaxbObj;
		try
		{
			/*
			 * TODO: support more efficient formats of XML content, e.g. gzipped XML, Fast Infoset, EXI.
			 */
			resultJaxbObj = xacmlParser.parse(xmlInputSrc);
		}
		catch (final JAXBException e)
		{
			throw new IndeterminateEvaluationException(
			        "PolicyProvider " + id + ": failed to parse Policy(Set) XML document from 'content' value of the policy document " + policyPOJO + " retrieved from database",
			        XacmlStatusCode.PROCESSING_ERROR.value(), e);
		}

		return new PolicyQueryResult(policyPOJO, resultJaxbObj, xacmlParser.getNamespacePrefixUriMap());
	}
