@Test
	public void testConformance() throws Exception
	{
		LOGGER.debug("Starting conformance test with files '{}*.xml'", testFilePathPrefix);

		final XmlnsFilteringParser respUnmarshaller = xacmlParserFactory.getInstance();
		final Response expectedResponse;
		final Path expectedRespFilepath = Paths.get(testFilePathPrefix + EXPECTED_RESPONSE_FILENAME_SUFFIX);
		if (Files.exists(expectedRespFilepath))
		{
			expectedResponse = TestUtils.createResponse(expectedRespFilepath, respUnmarshaller);
		}
		else
		{
			expectedResponse = null;
			// do nothing except logging -> request = null
			LOGGER.debug("Response file '{}' does not exist -> Static Policy/Request syntax error check", expectedRespFilepath);
		}

		final XmlnsFilteringParser reqUnmarshaller = xacmlParserFactory.getInstance();
		final Request request;
		// if no Request file, it is just a static policy syntax error check
		final Path reqFilepath = Paths.get(testFilePathPrefix + REQUEST_FILENAME_SUFFIX);
		if (Files.exists(reqFilepath))
		{
			try
			{
				request = TestUtils.createRequest(reqFilepath, reqUnmarshaller);
			}
			catch (final JAXBException e)
			{
				// we found syntax error in request
				if (expectedResponse == null)
				{
					// this is a Request syntax error check and we found the syntax error as
					// expected -> success
					LOGGER.debug("Successfully found syntax error as expected in Request located at: {}", reqFilepath);
					return;
				}

				// Unexpected error
				throw e;
			}
		}
		else
		{
			request = null;
			// do nothing except logging -> request = null
			LOGGER.debug("Request file '{}' does not exist -> Static policy syntax error check (Request/Response ignored)", reqFilepath);
		}

		/*
		 * Policies directory. If it exists, root Policy file is expected to be in there. This is the case for IIE*** conformance tests
		 */
		final Path policiesDir = Paths.get(testFilePathPrefix + POLICIES_DIRNAME_SUFFIX);

		final Path attributeProviderConfFile = Paths.get(testFilePathPrefix + ATTRIBUTE_PROVIDER_FILENAME_SUFFIX);
		final Optional<Path> optAttributeProviderConfFile = Files.isRegularFile(attributeProviderConfFile) ? Optional.of(attributeProviderConfFile) : Optional.empty();

		final PdpEngineConfiguration pdpEngineConf;
		try
		{
			if (Files.isDirectory(policiesDir))
			{
				final Path rootPolicyFile = policiesDir.resolve(ROOT_POLICY_FILENAME_SUFFIX);
				pdpEngineConf = TestUtils.newPdpEngineConfiguration(TestUtils.getPolicyRef(rootPolicyFile), policiesDir, enableXPath, optAttributeProviderConfFile, this.reqFilter, null);
			}
			else
			{
				final Path rootPolicyFile = Paths.get(testFilePathPrefix + ROOT_POLICY_FILENAME_SUFFIX);
				pdpEngineConf = TestUtils.newPdpEngineConfiguration(rootPolicyFile, enableXPath, optAttributeProviderConfFile, this.reqFilter, null);
			}
		}
		catch (final IllegalArgumentException e)
		{
			// we found syntax error in policy
			if (request == null)
			{
				// this is a policy syntax error check and we found the syntax error as
				// expected -> success
				LOGGER.debug("Successfully found syntax error as expected in policy(ies) with path: {}*", testFilePathPrefix);
				return;
			}

			// Unexpected error
			throw e;
		}

		try (PdpEngineInoutAdapter<Request, Response> pdp = PdpEngineAdapters.newXacmlJaxbInoutAdapter(pdpEngineConf))
		{
			if (request == null)
			{
				// this is a policy syntax error check and we didn't found the syntax error as
				// expected
				Assert.fail("Failed to find syntax error as expected in policy(ies)  with path: " + testFilePathPrefix + "*");
			}
			else if (expectedResponse == null)
			{
				/*
				 * No expected response, so it is not a PDP evaluation test, but request or policy syntax error check. We got here, so request and policy OK. This is unexpected.
				 */
				Assert.fail("Missing response file '" + expectedRespFilepath + "' or failed to find syntax error as expected in either request located at '" + reqFilepath
				        + "' or policy(ies) with path '" + testFilePathPrefix + "*'");

			}
			else
			{
				// this is an evaluation test with request/response (not a policy syntax check)
				LOGGER.debug("Request that is sent to the PDP: {}", request);
				final Response actualResponse = pdp.evaluate(request, reqUnmarshaller.getNamespacePrefixUriMap());
				if (LOGGER.isDebugEnabled())
				{
					LOGGER.debug("Response that is received from the PDP :  {}", TestUtils.printResponse(actualResponse));
				}

				TestUtils.assertNormalizedEquals(testFilePathPrefix, expectedResponse, actualResponse);
			}
		}
		catch (final IllegalArgumentException e)
		{
			// we found syntax error in policy
			if (request == null)
			{
				// this is a policy syntax error check and we found the syntax error as
				// expected -> success
				LOGGER.debug("Successfully found syntax error as expected in policy(ies) with path: {}*", testFilePathPrefix);
				return;
			}

			// Unexpected error
			throw e;
		}
	}
