public static void init(short sid,
                          String path,
                          LoggerFactory loggerFactory) throws Exception {
    name = new StringBuffer("AgentServer#").append(sid).toString();

    if (loggerFactory != null) Debug.setLoggerFactory(loggerFactory);
    logmon = Debug.getLogger(AgentServer.class.getName() + ".#" + sid);

    if (logmon.isLoggable(BasicLevel.DEBUG))
      logmon.log(BasicLevel.DEBUG, getName() + ", init()", new Exception());
    else
      logmon.log(BasicLevel.INFO, getName() + ", init()");
    
    synchronized(status) {
      if (status.value == Status.STOPPED) {
        logmon.log(BasicLevel.DEBUG, getName() + ", reset configuration");
        reset();
      }
      if (status.value != Status.INSTALLED)
        throw new Exception("cannot initialize, bad status: " + getStatusInfo());
      status.value = Status.INITIALIZING;
    }

    try {
      serverId = sid; 

      tgroup = new ThreadGroup(getName()) {
        public void uncaughtException(Thread t, Throwable e) {
          if (e instanceof VirtualMachineError) {
            if (logmon.isLoggable(BasicLevel.FATAL)) {
              logmon.log(BasicLevel.FATAL,
                         "Abnormal termination for " + t.getThreadGroup().getName() + "." + t.getName(), e);
              // AF: Should be AgentServer.stop() ?
              System.exit(-1);
            }
          } else {
            if (logmon.isLoggable(BasicLevel.WARN)) {
              logmon.log(BasicLevel.WARN,
                         "Abnormal termination for " + t.getThreadGroup().getName() + "." + t.getName(), e);
            }
          }
        }
      };
   
      //  Try to get transaction type from disk, then initialize the right
      // transaction manager and get the configuration.
      File dir = new File(path);
      if (dir.exists() && dir.isDirectory()) {
        File tfc = new File(dir, "TFC");
        if (tfc.exists()) {
          DataInputStream dis = null;
          try {
            dis = new DataInputStream(new FileInputStream(tfc));
            String tname = dis.readUTF();
            transaction = (Transaction) Class.forName(tname).newInstance();
          } catch (Exception exc) {
            logmon.log(BasicLevel.FATAL,
                       getName() + ", can't instantiate transaction manager", exc);
            throw new Exception("Can't instantiate transaction manager");
          } finally {
            if (dis != null) dis.close();
          }
          try {
            transaction.init(path);
          } catch (IOException exc) {
            logmon.log(BasicLevel.FATAL,
                       getName() + ", can't start transaction manager", exc);
            throw new Exception("Can't start transaction manager: " + exc.getMessage());
          }
        } else {
          // TODO (AF): We should probably return an exception as the TFC file does not
          // exist (normally the persistancy directory should be created by the transaction
          // initialization).
          logmon.log(BasicLevel.ERROR,
                     getName() + ", TFC file does not exist");
        }
      }

      // Gets static configuration of agent servers from a file. This method
      // fills the object graph configuration in the <code>A3CMLConfig</code>
      // object, then the configure method really initializes the server.
      // There are two steps because the configuration step needs the
      // transaction components to be initialized.
      if (transaction != null) {
        // Try to read the serialized configuration (through transaction)
        try {
          a3config = A3CMLConfig.load();
          logmon.log(BasicLevel.INFO, getName() + ", configuration retrieved from persistence");
        } catch (Exception exc) {
          if (logmon.isLoggable(BasicLevel.DEBUG))
            logmon.log(BasicLevel.DEBUG, getName() + ", config not found in persistence");
        }
        // Verify if the configuration has been updated since last boot.
        updatecfg = (ServerUpdate) transaction.load(UPDATE_CONF_TX_NAME);
        if (updatecfg != null) {
          logmon.log(BasicLevel.WARN,
                     getName() + ", configuration update detected: " + updatecfg.getUpdateDate());
          // Removes the object before the next boot
          transaction.delete(UPDATE_CONF_TX_NAME);
        }
      } else {
        // First boot, forces the update
        updatecfg = new ServerUpdate();
      }
      
      if (a3config == null) {
        //  Try to load an initial configuration (serialized or XML), or
        // generates a default one in case of failure.
        try {
          // Try to load a serialized configuration file directly in the filesystem.
          // TODO (AF): Is it used ? During Mediation deployment ?
          a3config = A3CMLConfig.getConfig(DEFAULT_SER_CFG_FILE);
          logmon.log(BasicLevel.INFO, getName() + ", serialized a3cmlconfig file loaded");
        } catch (Exception exc) {
          if (logmon.isLoggable(BasicLevel.DEBUG))
            logmon.log(BasicLevel.DEBUG, getName() + ", serialized a3cmlconfig file not found");
        }
      }
        
      if (a3config == null) {
        // Try to found an existing XML configuration file, then parse it.
        try {
          a3config = A3CML.getXMLConfig();
          logmon.log(BasicLevel.INFO, getName() + ", XML configuration file loaded");
        } catch (Exception exc) {
          if (logmon.isLoggable(BasicLevel.DEBUG))
            logmon.log(BasicLevel.DEBUG, getName() + ", XML configuration not found");
        }
      }

      if (a3config == null) {
        if (defaultConfig == null) {
          // If needed creates a Joram configuration as default.
          logmon.log(BasicLevel.WARN,
                     getName() + ", no configuration found, generates a default one");
          setDefaultConfig(sid);
        }

        // Try to parse default XML configuration.
        try {
          if (Log.logger.isLoggable(BasicLevel.DEBUG))
            Log.logger.log(BasicLevel.DEBUG,
                           getName() + ", start AgentServer with default configuration: \n" + defaultConfig);
          a3config = A3CML.getConfig(new StringReader(defaultConfig));
          logmon.log(BasicLevel.INFO,
                     getName() + ", use default XML configuration");
        } catch (Exception exc) {
          logmon.log(BasicLevel.WARN,
                     getName() + ", bad default XML configuration", exc);
        }
      }

      if (a3config == null) {
        // Last, Generate a basic A3CMLConfig.
        logmon.log(BasicLevel.WARN,
                   getName() + ", generates a default configuration");
        A3CMLDomain d = new A3CMLDomain(ADMIN_DOMAIN, SimpleNetwork.class.getName());
        A3CMLServer s = new A3CMLServer((short) sid, ADMIN_SERVER, "localhost");
        s.networks.addElement(new A3CMLNetwork(ADMIN_DOMAIN, 27300));
        d.addServer(s);
        a3config = new A3CMLConfig();
        a3config.addDomain(d);
        a3config.addServer(s);
      }

      // set properties
      setProperties(serverId);

      if (transaction == null) {
        try {
          String tname = getProperty("Transaction", "fr.dyade.aaa.ext.NGTransaction");
          transaction = (Transaction) Class.forName(tname).newInstance();
        } catch (Exception exc) {
          logmon.log(BasicLevel.FATAL, getName() + ", can't instantiate transaction manager", exc);
          throw new Exception("Can't instantiate transaction manager");
        }

        try {
          transaction.init(path);
        } catch (IOException exc) {
          logmon.log(BasicLevel.FATAL, getName() + ", can't start transaction manager", exc);
          throw new Exception("Can't start transaction manager");
        }
      }

      try {
        MXWrapper.registerMBean(transaction,
                                "AgentServer", "server=" + getName() + ",cons=Transaction");
      } catch (Exception exc) {
        if (logmon == null)
          logmon = Debug.getLogger(AgentServer.class.getName());
        logmon.log(BasicLevel.ERROR, getName() + " jmx failed", exc);
      }

      // save A3CMLConfig (Maybe we can omit it in some case).
      a3config.save();

      try {
        // Initialize AgentId class's variables.
        AgentId.init();
      } catch (ClassNotFoundException exc) {
        logmon.log(BasicLevel.FATAL,
                   getName() + ", can't initialize AgentId, bad classpath", exc);
        throw new Exception("Can't initialize AgentId, bad classpath");
      } catch (IOException exc) {
        logmon.log(BasicLevel.FATAL,
                   getName() + ", can't initialize AgentId", exc);
        throw new Exception("Can't initialize AgentId, storage problems");
      }

      try {
        // Configure the agent server.
        configure();
      } catch (Exception exc) {
        logmon.log(BasicLevel.FATAL, getName() + ", can't configure", exc);
        throw new Exception("Can't configure server: " + exc.getMessage());
      }

      try {
        // then restores all messages.
        String[] list = transaction.getList("@");
        for (int i=0; i<list.length; i++) {
          Message msg = Message.load(list[i]);

          if (msg.getSource() == serverId) {
            // The update has been locally generated, the message is ready to
            // deliver to its consumer (Engine or Network component). So we have
            // to insert it in the queue of this consumer.
            try {
              getServerDesc(msg.getDest()).getDomain().insert(msg);
            } catch (UnknownServerException exc) {
              logmon.log(BasicLevel.ERROR,
                         getName() + ", discard message to unknown server id#" +
                         msg.getDest());
              msg.delete();
              msg.free();
              continue;
            } catch (NullPointerException exc) {
              logmon.log(BasicLevel.ERROR,
                         getName() + ", discard message to unknown server id#" +
                         msg.getDest());
              msg.delete();
              msg.free();
              continue;
            } catch (ArrayIndexOutOfBoundsException exc) {
              logmon.log(BasicLevel.ERROR,
                         getName() + ", discard message to unknown server id#" +
                         msg.getDest());
              msg.delete();
              msg.free();
              continue;
            }
          } else {
            logmon.log(BasicLevel.ERROR,
                       getName() + ", discard undelivered message from server id#" +
                       msg.getDest());
            msg.delete();
            continue;
          }
        }
      } catch (ClassNotFoundException exc) {
        logmon.log(BasicLevel.FATAL,
                   getName() + ", can't restore messages", exc);
        throw new Exception("Can't restore messages, bad classpath");
      } catch (IOException exc) {
        logmon.log(BasicLevel.FATAL,
                   getName() + ", can't restore messages", exc);
        throw new Exception("Can't restore messages, storage problems");
      }

      // initializes channel before initializing fixed agents
      Channel.newInstance();    

      try {
        //  Initialize services, loads the configuration from transaction.
        ServiceManager.init();

        logmon.log(BasicLevel.INFO,
                   getName() + ", ServiceManager initialized");

        // Actually get Services from A3CML configuration file.
        // Be careful to not erase the ServiceManager list in order to avoid 
        // a new initialization of existing services.
        ServiceDesc services[] = AgentServer.getServices();
        if (services != null) {
          if (updatecfg != null) {
            logmon.log(BasicLevel.INFO,
                       getName() + ", Updates ServiceManager");
            
            Set<String> list = ServiceManager.listServices();
            for (int i = 0; i < services.length; i ++) {
              list.remove(services[i].getClassName());
              // Adds new services or validates existing ones.
              ServiceManager.register(services[i].getClassName(),
                                      services[i].getArguments());
            }
            for (String scname : list) {
              // Removes services no longer declared in configuration
              ServiceManager.unregister(scname);
            }
          }
        } else {
          logmon.log(BasicLevel.WARN,
                     getName() + ", no services from configuration");
        }
        ServiceManager.save();
      } catch (Exception exc) {
        logmon.log(BasicLevel.FATAL,
                   getName() + ", can't initialize services", exc);
        throw new Exception("Can't initialize services");
      }
      
      // Load the service classes before initializing the engine
      ServiceManager.loadServiceClasses();

      // initializes fixed agents
      engine.init(new AgentEngineContextImpl());

      logmon.log(BasicLevel.INFO,
                 getName() + ", initialized at " + new Date());

      // Commit all changes.
      transaction.begin();
      transaction.commit(true);

      try {
        SCServerMBean bean = new SCServer();
        MXWrapper.registerMBean(bean, "AgentServer", "server=" + getName());
      } catch (Exception exc) {
        if (logmon == null)
          logmon = Debug.getLogger(AgentServer.class.getName());
        logmon.log(BasicLevel.ERROR, getName() + " jmx failed", exc);
      }
    } catch (Exception exc) {
      logmon.log(BasicLevel.ERROR, getName() + "Cannot initialize", exc);
      synchronized(status) {
        // AF: Will be replaced by a BAD_INITIALIZED status allowing the
        // re-initialization..
        status.value = Status.INSTALLED;
      }
      throw exc;
    } catch (Throwable t) {
      logmon.log(BasicLevel.ERROR, getName() + "Cannot initialize", t);
      synchronized(status) {
        // AF: Will be replaced by a BAD_INITIALIZED status allowing the
        // re-initialization..
        status.value = Status.INSTALLED;
      }
      throw new Exception(t.getMessage());
    }

    long dterr = getLong(CFG_CHECK_PERIOD_PROPERTY, -1);
    if (dterr > 0) {
      checkServerTask = new CheckServerTask(dterr);
    }

    synchronized(status) {
      status.value = Status.INITIALIZED;
    }
  }
