synchronized void send(Destination dest, javax.jms.Message msg, int deliveryMode, int priority,
      long timeToLive, boolean timestampDisabled, long deliveryDelay, javax.jms.CompletionListener completionListener, MessageProducer messageProducer) throws JMSException {
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG,
                 "Session.send(" + dest + ',' + msg + ',' +
                 deliveryMode + ',' + priority + ',' + timeToLive + ',' + timestampDisabled + ',' + completionListener + ',' + messageProducer + ')');

    checkClosed();
    checkThreadOfControl();

    if (msg == null)
      throw new MessageFormatException("Cannot send null message");
    
    // Updating the message property fields:
    msg.setJMSMessageID(cnx.nextMessageId());
    msg.setJMSDeliveryMode(deliveryMode);
    msg.setJMSDestination(dest);
    
    if (timeToLive == 0) {
      msg.setJMSExpiration(0);
    } else {
      msg.setJMSExpiration(System.currentTimeMillis() + timeToLive);
    }
    msg.setJMSPriority(priority);
    if (!timestampDisabled) {
      msg.setJMSTimestamp(System.currentTimeMillis());
    }
    msg.setJMSRedelivered(false);

    // set the deliveryTime
    if (deliveryDelay > 0) {
      long deliveryTime = System.currentTimeMillis() + deliveryDelay;
      msg.setJMSDeliveryTime(deliveryTime);
    } else {
      msg.setJMSDeliveryTime(0L);
    }
    
    if (completionListener != null) {
      if (listener == null)
        listener = new CompletionListener(this, messageProducer);
      listener.addCompletionListener(completionListener, msg);
    }
    
    Message joramMsg = null;
    try {
      joramMsg = (Message) msg;
    } catch (ClassCastException exc) {
      try {
        // If the message to send is a non proprietary JMS message, try
        // to convert it.
        joramMsg = Message.convertJMSMessage(msg);
      } catch (JMSException jE) {
        MessageFormatException mE = new MessageFormatException("Message to send is invalid.");
        mE.setLinkedException(jE);
        throw mE;
      }
    }
    if (trace.isLoggable(BasicLevel.INFO))
      trace.log(BasicLevel.INFO,
                 this + " sending message=" + joramMsg + ", to=" + dest.getAdminName());
    //Add out interception...
    if ((outInterceptors != null) && (!outInterceptors.isEmpty())) {
      for (Iterator it = outInterceptors.iterator(); it.hasNext();) {
        MessageInterceptor interceptor = (MessageInterceptor) it.next();
        if (logger.isLoggable(BasicLevel.DEBUG))
          logger.log(BasicLevel.DEBUG, "Intercepting the message before sending by "
              + interceptor.getClass().getName());

        try {
          interceptor.handle(joramMsg, this);
        } catch (Throwable t) {
          if (logger.isLoggable(BasicLevel.WARN))
            logger.log(BasicLevel.WARN, "Warning while interception (continue anyway...)", t);
        }
      }
    }
    joramMsg.prepare();

    //set clientID
    joramMsg.momMsg.clientID = cnx.getClientID();
    
    if (transacted) {
      if (logger.isLoggable(BasicLevel.DEBUG))
        logger.log(BasicLevel.DEBUG, "Buffering the message.");
      // If the session is transacted, keeping the request for later delivery:
      prepareSend(dest, (org.objectweb.joram.shared.messages.Message) joramMsg.momMsg.clone());
    } else {
      ProducerMessages pM = new ProducerMessages(dest.getName(),
          (org.objectweb.joram.shared.messages.Message) joramMsg.momMsg.clone());

      if (logger.isLoggable(BasicLevel.DEBUG))
        logger.log(BasicLevel.DEBUG, "Sending " + joramMsg);

      if (asyncSend || (!joramMsg.momMsg.persistent)) {
        // Asynchronous sending
        pM.setAsyncSend(true);
        mtpx.sendRequest(pM, listener);
      } else {
        requestor.request(pM, listener);
      }
      listener = null;
    }
  }
