private void initialize(boolean firstTime) throws Exception {
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "--- " + this + " (re)initializing...");
    
    topicsTable = new Hashtable();
    
    if (firstTime) {
      arrivalState = new UserAgentArrivalState(ARRIVAL_STATE_PREFIX + getId().toString());
    } else {
      arrivalState = UserAgentArrivalState.load(ARRIVAL_STATE_PREFIX + getId().toString());
    }
     
    MessageTableFactory messageTableFactory = MessageTableFactory.newFactory();
    messagesTable = messageTableFactory.createMessageTable(MESSAGE_TABLE_PREFIX + getId().toString());

    if (contexts == null) contexts = new Hashtable<Integer, ClientContext>();
    if (subsTable == null) subsTable = new Hashtable();
    if (subsClientIDs == null) subsClientIDs = new Properties();

    Transaction tx = AgentServer.getTransaction();
    String[] persistedClientNames = tx.getList(ClientContext.getTransactionPrefix(getId()));
    for (int i = 0; i < persistedClientNames.length; i++) {
      logger.log(BasicLevel.INFO, "ClientContext named [" + persistedClientNames[i] + "] loaded");
      
      try {
        ClientContext cc = (ClientContext) tx.load(persistedClientNames[i]);
        cc.txName = persistedClientNames[i];
        cc.setProxyId(getId());
        cc.setProxyAgent(this);
        contexts.put(cc.getId(), cc);
        addValidCC(cc);
      } catch (Exception exc) {
        logger.log(BasicLevel.ERROR, 
                   "ClientContext named [" + persistedClientNames[i] + "] could not be loaded", exc);
      }
    }

    String[] persistedSubscriptionNames = tx.getList(ClientSubscription.getTransactionPrefix(getId()));
    for (int i = 0; i < persistedSubscriptionNames.length; i++) {
      logger.log(BasicLevel.INFO, "ClientSubscription named [" + persistedSubscriptionNames[i] + "] loaded");

      try {
        ClientSubscription cs = (ClientSubscription) tx.load(persistedSubscriptionNames[i]);
        cs.txName = persistedSubscriptionNames[i];
        cs.setProxyId(getId());
        cs.setProxyAgent(this);
        
        cs.initMessageIds();
        cs.loadMessageIds();
        
        subsTable.put(cs.getName(), cs);
      } catch (Exception exc) {
        logger.log(BasicLevel.ERROR,
                   "ClientSubscription named [" + persistedSubscriptionNames[i] + "] could not be loaded", exc);
      }
    }

    setActiveCtxId(-1);

    // Re-initializing after a crash or a server stop.

    // interceptors
    if (interceptorsPropOUT != null) {
      interceptorsOUT = new ArrayList<MessageInterceptor>();
    	InterceptorsHelper.addInterceptors(getAgentId(), getName(), AdminCommandConstant.INTERCEPTORS_OUT, interceptorsPropOUT, interceptorsOUT);
    }
    if (interceptorsPropIN != null) {
      interceptorsIN = new ArrayList<MessageInterceptor>();
    	InterceptorsHelper.addInterceptors(getAgentId(), getName(), AdminCommandConstant.INTERCEPTORS_IN, interceptorsPropIN, interceptorsIN);
    }

    // Browsing the pre-crash contexts:
    ClientContext activeCtx;
    AgentId destId;
    for (Iterator ctxs = contexts.values().iterator(); ctxs.hasNext();) {
      activeCtx = (ClientContext) ctxs.next();
      activeCtx.setProxyAgent(this);
      ctxs.remove();
      
      // Denying the non acknowledged messages:
      for (Iterator queueIds = activeCtx.getDeliveringQueues(); queueIds.hasNext();) {
        destId = (AgentId) queueIds.next();
        sendNot(destId, new DenyRequest(activeCtx.getId()));

        if (logger.isLoggable(BasicLevel.DEBUG))
          logger.log(BasicLevel.DEBUG, "Denies messages on queue " + destId.toString());
      }

      // Saving the prepared transactions.
      Iterator xids = activeCtx.getTxIds();
      Xid xid;
      XACnxPrepare recoveredPrepare;
      XACnxPrepare prepare;
      while (xids.hasNext()) {
        if (recoveredTransactions == null)
          recoveredTransactions = new Hashtable();

        xid = (Xid) xids.next();

        recoveredPrepare = (XACnxPrepare) recoveredTransactions.get(xid);
        prepare = activeCtx.getTxPrepare(xid);

        if (recoveredPrepare == null)
          recoveredTransactions.put(xid, prepare);
        else {
          recoveredPrepare.getSendings().addAll(prepare.getSendings());
          recoveredPrepare.getAcks().addAll(prepare.getAcks());
        }
      }

      // Deleting the temporary destinations:
      for (Iterator tempDests = activeCtx.getTempDestinations(); tempDests.hasNext();) {
        destId = (AgentId) tempDests.next();
        deleteTemporaryDestination(destId);

        if (logger.isLoggable(BasicLevel.DEBUG))
          logger.log(BasicLevel.DEBUG, "Deletes temporary destination " + destId.toString());
      }
      
      // destroy the context
      logger.log(BasicLevel.INFO, "ClientContext named [" + activeCtx.txName + "] deleted");
      activeCtx.delete();
    }

    // Retrieving the subscriptions' messages.
    List messages = Message.loadAll(getMsgTxname(), Integer.MAX_VALUE);
    
    if (subsTable.isEmpty()) {
      // it is possible because we always save MessageSoftRef
      // so we must delete all message.
      Message.deleteAll(getMsgTxname());
    }

    // Browsing the pre-crash subscriptions:
    List topics = new ArrayList();
    for (Iterator subs = subsTable.entrySet().iterator(); subs.hasNext();) {
      Map.Entry subEntry = (Map.Entry) subs.next();
      ClientSubscription cSub = (ClientSubscription) subEntry.getValue();
      destId = cSub.getTopicId();
      if (!topics.contains(destId))
        topics.add(destId);
      // Deleting the non durable subscriptions.
      if (!cSub.getDurable()) {
        // TODO (AF): Normally the non durable subscription are not saved (see ClientSubscription).
        subs.remove();
        try {
          MXWrapper.unregisterMBean(getSubMBeanName((String) subEntry.getKey()));
        } catch (Exception e1) {
          if (logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "  - Problem when unregistering ClientSubscriptionMbean", e1);
        }
        // TODO (AF): we should remove it using cSub.delete() but it is useless as it is impossible.
      } else {
      // Reinitializing the durable ones.
        cSub.setProxyAgent(this);
        cSub.reinitialize(messagesTable, messages, true);
        try {
          MXWrapper.registerMBean(cSub, getSubMBeanName((String) subEntry.getKey()));
        } catch (Exception e1) {
          if (logger.isLoggable(BasicLevel.WARN))
            logger.log(BasicLevel.WARN, "  - Could not register ClientSubscriptionMbean", e1);
        }
        TopicSubscription tSub = (TopicSubscription) topicsTable.get(destId);
        if (tSub == null) {
          tSub = new TopicSubscription();
          topicsTable.put(destId, tSub);
        }
        tSub.putSubscription((String) subEntry.getKey(), cSub.getSelector());
      }
    }
    // Browsing the topics and updating their subscriptions.
    for (Iterator topicIds = topics.iterator(); topicIds.hasNext();)
      updateSubscriptionToTopic((AgentId) topicIds.next(), -1, -1);
  }
