private void acceptConnection() throws Exception {
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "TcpConnectionListener.acceptConnection()");

    Socket sock = proxyService.getServerSocket().accept();
    String inaddr = sock.getInetAddress().getHostAddress();

    connectionCount++;

    if (logger.isLoggable(BasicLevel.INFO))
      logger.log(BasicLevel.INFO, " -> accept connection from " + inaddr);

    try {
      sock.setTcpNoDelay(true);

      // Fix bug when the client doesn't use the right protocol (e.g. Telnet)
      // and blocks this listener.
      sock.setSoTimeout(timeout);

      InputStream is = sock.getInputStream();
      NetOutputStream nos = new NetOutputStream(sock);

      byte[] magic = null;
      try {
        magic = StreamUtil.readByteArrayFrom(is, 8);
      } catch (IOException exc) {
        String errorMsg = "Connection from " + sock.getInetAddress() + ':' + sock.getPort() + " cannot read magic number. Client is not compatible with JORAM.";
        protocolErrorCount++;
        throw new IllegalAccessException(errorMsg);
      }
      
      for (int i = 0; i < 5; i++) {
        if (magic.length == i || magic[i] != MetaData.joramMagic[i] && magic[i] > 0) {
          String errorMsg = "Connection from " + sock.getInetAddress() + ':' + sock.getPort() + " bad magic number. Client is not compatible with JORAM.";
          protocolErrorCount++;
          throw new IllegalAccessException(errorMsg);
        }
      }
      if (magic[7] != MetaData.joramMagic[7]) {
        if (magic[7] > 0 && MetaData.joramMagic[7] > 0) {
          String errorMsg = "Connection from " + sock.getInetAddress() + ':' + sock.getPort() + " bad protocol version number " + magic[7] + " != " + MetaData.joramMagic[7];
          protocolErrorCount++;
          throw new IllegalAccessException(errorMsg);
        }
        
        logger.log(BasicLevel.WARN,
                   "Wildcard protocol version number: from stream = " + magic[7] + ", from MetaData = " + MetaData.joramMagic[7]);
      }

      //read the ack mode(noAckedQueue)
      boolean noAckedQueue = StreamUtil.readBooleanFrom(is);
      if (logger.isLoggable(BasicLevel.DEBUG))
        logger.log(BasicLevel.DEBUG, " -> read noAckedQueue = " + noAckedQueue);
      if (noAckedQueue)
        TcpProxyService.createExecutors();
      
      long dt = Math.abs(StreamUtil.readLongFrom(is) - System.currentTimeMillis());
      if (dt > clockSynchroThreshold)
        logger.log(BasicLevel.WARN, " -> bad clock synchronization between client and server: " + dt);
      StreamUtil.writeTo(dt, nos);

      Identity identity = Identity.read(is);
      if (logger.isLoggable(BasicLevel.DEBUG))
        logger.log(BasicLevel.DEBUG, " -> read identity = " + identity);

      int key = StreamUtil.readIntFrom(is);
      if (logger.isLoggable(BasicLevel.DEBUG))
        logger.log(BasicLevel.DEBUG, " -> read key = " + key);

      int heartbeat = 0;
      if (key == -1) {
        heartbeat = StreamUtil.readIntFrom(is);
        if (logger.isLoggable(BasicLevel.DEBUG))
          logger.log(BasicLevel.DEBUG, " -> read heartBeat = " + heartbeat);
      }
      
      if (logger.isLoggable(BasicLevel.INFO))
        logger.log(BasicLevel.INFO, " -> open connection " + identity + "/" + key + " - " + heartbeat);
      
      GetProxyIdNot gpin = new GetProxyIdNot(identity, inaddr);
      AgentId proxyId;
      try {
        gpin.invoke(AdminTopic.getDefault(), connectingTimer);
        proxyId = gpin.getProxyId();
        if (proxyId == null) {
          String msg = "JMS connection cancelled after " + (connectingTimer /1000) + " secs.";
          IllegalStateException jmsExc = new IllegalStateException(msg);
          
          if (logger.isLoggable(BasicLevel.WARN))
            logger.log(BasicLevel.WARN,
                       " -> JMS connection cancelled after " + (connectingTimer /1000) + " secs.", jmsExc);
          
          throw jmsExc;
        }
      } catch (Exception exc) {
        if (logger.isLoggable(BasicLevel.WARN))
          logger.log(BasicLevel.WARN, " -> login failed", exc);
        failedLoginCount++;
        StreamUtil.writeTo(1, nos);
        StreamUtil.writeTo(exc.getMessage(), nos);
        nos.send();
        return;
      }

      if (logger.isLoggable(BasicLevel.DEBUG))
        logger.log(BasicLevel.DEBUG, " -> open connection, proxyId=" + proxyId);
      
      IOControl ioctrl;
      ReliableConnectionContext ctx;
      if (key == -1) {
        OpenConnectionNot ocn = new OpenConnectionNot(Type.RELIABLE, heartbeat, noAckedQueue);
        try {
          ocn.invoke(proxyId, connectingTimer);
          ctx = (ReliableConnectionContext) ocn.getConnectionContext();
          if (ctx == null) {
            String msg = "JMS connection cancelled after " + (connectingTimer /1000) + " secs.";
            IllegalStateException jmsExc = new IllegalStateException(msg);

            if (logger.isLoggable(BasicLevel.WARN))
              logger.log(BasicLevel.WARN,
                         "JMS connection cancelled after " + (connectingTimer /1000) + " secs.", jmsExc);
            
            // Note: The probability that the notification has been processed by the user proxy is
            // very small. In this case we have only to clean the context object in the proxy's connection
            // table. In the best case this item will be cleaned up if the heartbeat is activated. In the
            // worst case it will be cleaned at restart.
            // If we really want to treat this case it would be necessary to issue a CloseConnectionNot with
            // the connection key.

            // TODO (AF): We have to clean the pending connection if any.

            throw jmsExc;
          }
        } catch (Exception exc) {
          if (logger.isLoggable(BasicLevel.WARN))
            logger.log(BasicLevel.WARN, " -> login failed", exc);
          failedLoginCount++;
          StreamUtil.writeTo(1, nos);
          StreamUtil.writeTo(exc.getMessage(), nos);
          nos.send();
          return;
        }

        StreamUtil.writeTo(0, nos);
        key = ctx.getKey();
        StreamUtil.writeTo(key, nos);
        nos.send();
        ioctrl = new IOControl(sock);
      } else {
        GetConnectionNot gcn = new GetConnectionNot(key);
        try {
          gcn.invoke(proxyId);
        } catch (Exception exc) {
          if (logger.isLoggable(BasicLevel.WARN))
            logger.log(BasicLevel.WARN, "TcpConnectionListener: reconnection failed", exc);
          StreamUtil.writeTo(1, nos);
          StreamUtil.writeTo(exc.getMessage(), nos);
          nos.send();
          return;
        }
        ctx = (ReliableConnectionContext) gcn.getConnectionContext();
        StreamUtil.writeTo(0, nos);
        nos.send();
        ioctrl = new IOControl(sock, ctx.getInputCounter());

        // Close the remaining connection if it exists
        TcpConnection tcpConnection = proxyService.getConnection(proxyId, key);
        if (tcpConnection != null) {
          tcpConnection.close();
        }
      }

      // Reset the timeout in order to enable the server to indefinitely
      // wait for requests.
      sock.setSoTimeout(0);

      TcpConnection tcpConnection = new TcpConnection(ioctrl, ctx, proxyId, proxyService, identity);
      tcpConnection.start();
    } catch (IllegalAccessException exc) {
      if (logger.isLoggable(BasicLevel.DEBUG))
        logger.log(BasicLevel.ERROR, "TcpConnectionListener: close connection", exc);
      else
        logger.log(BasicLevel.ERROR, "TcpConnectionListener: close connection, " + exc.getMessage());
      sock.close();
      throw exc;
    } catch (IOException exc) {
      if (logger.isLoggable(BasicLevel.DEBUG))
        logger.log(BasicLevel.WARN, "TcpConnectionListener: close connection", exc);
      else
        logger.log(BasicLevel.WARN, "TcpConnectionListener: close connection, " + exc.getMessage());
      sock.close();
      throw exc;
    }
  }
