public void distribute(Message message) throws Exception {
  	if (logger.isLoggable(BasicLevel.DEBUG)) {
      logger.log(BasicLevel.DEBUG, "JMSDistribution.distribute(" + message + ')');
  	}
  	
    List<String> connectionNames = this.connectionNames;
    if (message.properties != null) {
      Object customRouting = message.properties.get(ROUTING_PROP);
      if (customRouting != null && customRouting instanceof String) {
        connectionNames = JMSConnectionService.convertToList((String) customRouting);
      }
    }
    if (logger.isLoggable(BasicLevel.DEBUG)) {
      logger.log(BasicLevel.DEBUG, "JMSDistribution.distribute: connectionNames = " + connectionNames);
    }
      
    // Update sessions if necessary
    long now = System.currentTimeMillis();
    if (now - lastUpdate > updatePeriod) {
      if (logger.isLoggable(BasicLevel.DEBUG)) {
        logger.log(BasicLevel.DEBUG, "Updating sessions.");
      }
      List<JMSModule> connections = JMSConnectionService.getInstance().getConnections();
      if (logger.isLoggable(BasicLevel.DEBUG)) {
        logger.log(BasicLevel.DEBUG, "JMSDistribution.distribute: connections=" + connections.size());
      } else if (logger.isLoggable(BasicLevel.INFO) && connections.isEmpty() && ((now - lastwarn) > 60000L)) {
        logger.log(BasicLevel.INFO, "JMSDistribution.distribute: no available connection");
        lastwarn = now;
      }
            
      for (final JMSModule connection : connections) {
        SessionAndProducer sap = sessions.get(connection.getName());
        if (sap != null) {
          // Verify that the connection still valid
          if (sap.connection != connection.getCnx())  {
            if (logger.isLoggable(BasicLevel.INFO))
              logger.log(BasicLevel.INFO,
                         "JMSDistribution.distribute: remove outdated connection " + connection.getName());
            sessions.remove(connection.getName());
            sap = null;
          }
        }
        
        if (sap == null) { // !sessions.containsKey(connection.getCnxFactName()))
          if (logger.isLoggable(BasicLevel.INFO))
            logger.log(BasicLevel.INFO,
                       "JMSDistribution.distribute: Creates new connection for distribution, cf = " + connection.getName());
          try {
            Session session = connection.getCnx().createSession(false, Session.AUTO_ACKNOWLEDGE);
            Destination dest = (Destination) connection.retrieveJndiObject(destName);
            MessageProducer producer = session.createProducer(dest);
            sessions.put(connection.getName(), new SessionAndProducer(connection.getCnx(), session, producer));
            if (logger.isLoggable(BasicLevel.INFO))
              logger.log(BasicLevel.INFO, "JMSDistribution.distribute: New connection available.");
          } catch (Exception exc) {
            if (logger.isLoggable(BasicLevel.INFO)) {
              logger.log(BasicLevel.INFO, "Connection still not usable.", exc);
            }
          }
        }
      }
      lastUpdate = now;
    }

    // Send the message
    Iterator<Map.Entry<String, SessionAndProducer>> iter = sessions.entrySet().iterator();
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "Try to send message, sessions available: " + sessions.size());
    while (iter.hasNext()) {
      Map.Entry<String, SessionAndProducer> entry = iter.next();
      try {
        SessionAndProducer session = entry.getValue();
        String cnxName = entry.getKey();
        if (logger.isLoggable(BasicLevel.DEBUG))
          logger.log(BasicLevel.DEBUG, "Try session: " + cnxName);
        if (connectionNames != null && !connectionNames.contains(cnxName)) {
          continue;
        }
        if (logger.isLoggable(BasicLevel.DEBUG)) {
          logger.log(BasicLevel.DEBUG, "Sending message using " + cnxName);
        }
        if (session.isJoramSession()) {
        	// convert a Joram message because this message is modified on session send.
        	// And if we have an exception we must keep the original message.
        	session.producer.send(org.objectweb.joram.client.jms.Message.convertJMSMessage(org.objectweb.joram.client.jms.Message.wrapMomMessage(null, message)));
        } else {
          // JORAM-308: It seems that the original message could be modified by the provider even if this one is 
          // not Joram. So we have to create a new message from the target session and initialize it from the
          // original message.
        	session.producer.send(org.objectweb.joram.client.jms.Message.convertJMSMessage(org.objectweb.joram.client.jms.Message.wrapMomMessage(null, message)));
        }
        sessions.get(cnxName); // Access the used connection to update the LRU map
        return;
      } catch (JMSException exc) {
        if (logger.isLoggable(BasicLevel.WARN)) {
          logger.log(BasicLevel.WARN, "Session is not usable, remove from table.", exc);
        }
        // TODO (AF): We should indicate that the corresponding connection is no longer
        // available calling JMSModule.onException().
        iter.remove();
      }
    }

    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG,
                 "Message could not be sent, no usable connection/session found - " + sessions.size());

    throw new Exception("Message could not be sent, no usable connection/session found - " + sessions.size());
  }
