@GET
  @Path("/{name}")
  @Produces({MediaType.TEXT_PLAIN, MediaType.APPLICATION_JSON})
  @Consumes(MediaType.TEXT_PLAIN)
  public Response consumeMsg(
      @Context HttpHeaders headers,
      @PathParam("name") String consName,
      @DefaultValue("-1")@QueryParam("timeout") long timeout,
      @DefaultValue("false")@QueryParam("no-local") boolean noLocal,
      @DefaultValue("false")@QueryParam("durable") boolean durable,
      @DefaultValue("false")@QueryParam("shared") boolean shared,
      @QueryParam("sub-name") String subName,
      @Context UriInfo uriInfo) {

    if (logger.isLoggable(BasicLevel.INFO))
      logger.log(BasicLevel.INFO, "GET: " + uriInfo.getAbsolutePathBuilder());
    
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "consumeTextMsg(" + headers + ", " + consName + ", " + timeout + ", " +
          noLocal + ", " + durable + ", " + shared + ", " + subName + ", " + uriInfo);

    Response.ResponseBuilder builder = null;
    try {
      if (consName == null) {
        if (logger.isLoggable(BasicLevel.WARN))
          logger.log(BasicLevel.WARN, "sendMsg: The consumer name is null.");
        builder = Response.status(Response.Status.EXPECTATION_FAILED).entity("The consumer name is null.");
        return builder.build();
      }

      Message message = null;
      try {
        // receive the message
        message = helper.consume(consName, timeout, noLocal, durable, shared, subName, -1);
        if (logger.isLoggable(BasicLevel.DEBUG))
          logger.log(BasicLevel.DEBUG, "consumeMsg: message = " + message); 
      } catch (Exception e) {
        if (logger.isLoggable(BasicLevel.WARN))
          logger.log(BasicLevel.WARN, "", e);
        builder = Response.status(Response.Status.INTERNAL_SERVER_ERROR).entity(e.getMessage());
        return builder.build();
      }
      
      Object msg = null;
      boolean jsonMedia = false;
      List<MediaType> medias = headers.getAcceptableMediaTypes();
      if (medias.contains(MediaType.APPLICATION_JSON_TYPE)) {
        jsonMedia = true;
      }
      
      try {
        if (message instanceof TextMessage) {
          if (jsonMedia) {
            HashMap jsonMap = new HashMap<>();
            jsonMap.put(JMS_TYPE, message.getClass().getSimpleName());
            jsonMap.put(JMS_BODY, ((TextMessage) message).getText());
            jsonMap.put(JMS_PROPERTIES, getPropertiesToJsonMap(message));
            jsonMap.put(JMS_HEADER, getHeaderToJsonMap(message));
            msg = jsonMap;
          } else {
            msg = ((TextMessage) message).getText();
          }
          if (logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "consumeTextMsg: msg = " + msg);

        } else if (message instanceof MapMessage) {
          if (jsonMedia) {
            HashMap jsonMap = new HashMap<>();
            jsonMap.put(JMS_TYPE, message.getClass().getSimpleName());
            jsonMap.put(JMS_BODY, getMapMessageToJsonBodyMap(message));
            jsonMap.put(JMS_PROPERTIES, getPropertiesToJsonMap(message));
            jsonMap.put(JMS_HEADER, getHeaderToJsonMap(message));
            msg = jsonMap;
          } else {
            builder = Response.status(Response.Status.NO_CONTENT).entity("Only available with MediaType.APPLICATION_JSON");
            return builder.build();
          }

        } else if (message instanceof BytesMessage) {
          if (jsonMedia) {
            HashMap jsonMap = new HashMap<>();
            jsonMap.put(JMS_TYPE, message.getClass().getSimpleName());
            jsonMap.put(JMS_BODY, message.getBody(byte[].class));
            jsonMap.put(JMS_PROPERTIES, getPropertiesToJsonMap(message));
            jsonMap.put(JMS_HEADER, getHeaderToJsonMap(message));
            msg = jsonMap;
          } else {
            builder = Response.status(Response.Status.NO_CONTENT).entity("Only available with MediaType.APPLICATION_JSON");
            return builder.build();
          }
          
        } else if (message != null) {
          throw new JMSException("Invalide Message type: " + message);
        }
      } catch (JMSException e) {
        if (logger.isLoggable(BasicLevel.WARN))
          logger.log(BasicLevel.WARN, "", e);
        builder = Response.status(Response.Status.INTERNAL_SERVER_ERROR).entity(e.getMessage());
        return builder.build();
      }
      
      if (timeout > 0 && message == null) {
        // time out expire
        builder = Response.status(Response.Status.NO_CONTENT);
      } else {
        // add message entity for the client
        if (jsonMedia) {
          Gson gson = new GsonBuilder().create();
          String json = gson.toJson(msg);
          if (logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "json = " + json);
          builder = Response.status(Response.Status.OK).type(MediaType.APPLICATION_JSON_TYPE).entity(json);
        } else {
          if (logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG, "msg = " + msg);
          builder = Response.status(Response.Status.OK).entity(msg);
        }
      }
      
      ConsumerContext consCtx = (ConsumerContext) helper.getSessionCtx(consName);
      try {
        if (consCtx.getJmsContext().getTransacted()) {
          // link commit consumer message
          UriBuilder nextBuilder = uriInfo.getAbsolutePathBuilder().path(CONTEXT_COMMIT);
          builder.link(nextBuilder.build(), CONTEXT_COMMIT);
          
          // link rollback consumer message
          nextBuilder = uriInfo.getAbsolutePathBuilder().path(CONTEXT_ROLLBACK);
          builder.link(nextBuilder.build(), CONTEXT_ROLLBACK);
        }
      } catch (Exception e) {
        if (logger.isLoggable(BasicLevel.WARN))
          logger.log(BasicLevel.WARN, "", e);
        builder = Response.status(Response.Status.INTERNAL_SERVER_ERROR).entity(e.toString());
        return builder.build();
      }

      try {
        if (consCtx.getJmsContext().getSessionMode() == JMSContext.CLIENT_ACKNOWLEDGE) {
          // link acknowledge all message
          UriBuilder nextBuilder = uriInfo.getAbsolutePathBuilder();
          builder.link(nextBuilder.build(), CONTEXT_ACK);

          long id = consCtx.getId(message);
          if (id > 0) {
            // link acknowledge a message
            nextBuilder = uriInfo.getAbsolutePathBuilder().path(""+id);
            builder.link(nextBuilder.build(), CONTEXT_ACK_MSG);
          }
        }

        long id = consCtx.getLastId();
        // link consume next message
        UriBuilder nextBuilder = uriInfo.getAbsolutePathBuilder().path(""+(id+1));
        builder.link(nextBuilder.build(), CONTEXT_CONSUME_NEXT);

      } catch (Exception e) {
        if (logger.isLoggable(BasicLevel.WARN))
          logger.log(BasicLevel.WARN, "", e);
        builder = Response.status(Response.Status.INTERNAL_SERVER_ERROR).entity(e.toString());
        return builder.build();
      }

      // link consume message
      UriBuilder nextBuilder = uriInfo.getAbsolutePathBuilder();
      URI next = nextBuilder.build();
      builder.link(next, CONTEXT_CONSUME);

      return builder.build();
    } finally {
      logLinks(builder);
    }
  }
