/*
 * JORAM: Java(TM) Open Reliable Asynchronous Messaging
 * Copyright (C) 2006 - 2020 ScalAgent Distributed Technologies
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): ScalAgent Distributed Technologies
 * Contributor(s): 
 */
package org.objectweb.joram.mom.messages;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.lang.ref.SoftReference;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import org.objectweb.joram.mom.util.JoramHelper;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import fr.dyade.aaa.agent.AgentServer;
import fr.dyade.aaa.common.Debug;
import fr.dyade.aaa.common.encoding.Decoder;
import fr.dyade.aaa.common.encoding.Encodable;
import fr.dyade.aaa.common.encoding.EncodableFactory;
import fr.dyade.aaa.common.encoding.Encoder;
import fr.dyade.aaa.common.stream.StreamUtil;
import fr.dyade.aaa.util.Transaction;

/** 
 * The <code>Message</code> class actually provides the transport facility
 * for the data exchanged during MOM operations.
 * <p>
 * A message content is always wrapped as a bytes array, it is characterized
 * by properties and "header" fields.
 */
public final class Message implements Comparable<Message>, Serializable, Encodable, MessageView {
  /** define serialVersionUID for interoperability */
  private static final long serialVersionUID = 2L;

  /** logger */
  public static Logger logger = Debug.getLogger(Message.class.getName());

  /** Arrival position of this message on its queue or proxy. */
  transient public long order;

  @Override
  public int compareTo(Message msg) {
    // This method allows to sort messages list in the history order (see JORAM-358).
    // Be careful, specific comparators are using in loadAll method.
    return Long.compare(order, msg.order);
  }

  /**
   * The number of acknowledgements a message still expects from its 
   * subscribers before having been fully consumed by them (field used
   * by JMS proxies).
   * Be careful, this field is not saved but set to 0 during message
   * loading then calculated during the proxy initialization.
   */
  public transient int acksCounter;

  /**
   * The number of acknowledgements a message still expects from its 
   * durable subscribers before having been fully consumed by them (field used
   * by JMS proxies).
   * Be careful, this field is not saved but set to 0 during message
   * loading then calculated during the proxy initialization.
   */
  public transient int durableAcksCounter;

  /**
   * Reference to the MOM message.
   * Should not be used directly, rather use getMsg so the message is restored from
   * disk if necessary (currently not used)
   */
  private transient org.objectweb.joram.shared.messages.Message msg;
  
  public /*@ pure @*/ final org.objectweb.joram.shared.messages.Message getMsg() {
    // TODO (AF): If the message is not loaded, load it from disk (JORAM-358).
    // Currently should never happened.
    if (msg == null) {
      try {
        Message m = load(txname);
        msg = m.msg;
      } catch (ClassNotFoundException | IOException exc) {
        logger.log(BasicLevel.ERROR, "Cannot load message " + txname, exc);
        // TODO (AF): Should remove this message in the list?
      }
    }
    return msg;
  }

  /**
   *  Defines if the queue restoration at startup must use the loadAll transaction
   * mechanism when it is implemented. 
   * <p>
   *  Default value is false.
   * <p>
   *  This property can be fixed either from <code>java</code> launching
   * command, or in <code>a3servers.xml</code> configuration file.
   */
  private static final boolean useLoadALL =
      AgentServer.getBoolean("org.objectweb.joram.mom.messages.USELOADALL");
  
  /** SoftReference to the body of the MOM message. */
  private transient SoftReference bodySoftRef = null;

  /** <code>true</code> if soft reference is used for the message. */
  private transient boolean soft;

  /**
   *  Defines if the swapping mechanism is globally activated for messages
   * in this server.
   * <p>
   *  Default value is false.
   * <p>
   *  Note: the message swapping can be finely configured using the 
   * <code>JMS_JORAM_SWAPALLOWED</code> property of the JMS message.
   * <p>
   *  This property can be fixed either from <code>java</code> launching
   * command, or in <code>a3servers.xml</code> configuration file.
   */
  private static final boolean globalUseSoftRef =
    AgentServer.getBoolean("org.objectweb.joram.mom.messages.SWAPALLOWED");
  
  /**
   * Empty constructor.
   */
  public Message() {}

  /**
   * Constructs a <code>Message</code> instance.
   */
  public Message(org.objectweb.joram.shared.messages.Message msg) {
    this.msg = msg;
    // Soft reference can be used only if message is persistent and has a body.
    Boolean msgUseSoftRef = (Boolean) msg.getProperty("JMS_JORAM_SWAPALLOWED");
    if (msgUseSoftRef != null) {
      this.soft = msgUseSoftRef.booleanValue() && msg.persistent && msg.body != null;
    } else {
      this.soft = globalUseSoftRef && msg.persistent && msg.body != null;
    }
  }

  /**
   * Returns the contained message eventually without the body.
   * 
   * @return The contained message.
   */
  public org.objectweb.joram.shared.messages.Message getHeaderMessage() {
    // Loads the message if needed.
    getMsg();
    
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "MessagePersistenceModule.getHeaderMessage() -> " + msg);
    
    return msg;
  }

  /**
   * Returns the contained message with body.
   * If needed the body is loaded from repository.
   * 
   * @return The contained message.
   */
  public org.objectweb.joram.shared.messages.Message getFullMessage() {
    getMsg();
    
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "MessagePersistenceModule.getFullMessage() " + txname);
    // The message can be soft but releaseFullMessage has not been called !
    if (!soft || msg.body != null)
      return msg;

    if (bodySoftRef != null) {
      msg.body = (byte[]) bodySoftRef.get();
      bodySoftRef = null;
      if (msg.body != null) {
        return msg;
      }
    }

    // Try to load the body from repository
    try {
      msg.body = AgentServer.getTransaction().loadByteArray(txname + "B");
      if (logger.isLoggable(BasicLevel.DEBUG))
        logger.log(BasicLevel.DEBUG, "Body loaded.");
      bodySoftRef = null;
    } catch (Exception exc) {
      logger.log(BasicLevel.ERROR,
                 "Body of message named [" + txname + "] could not be loaded", exc);
    }
    
    return msg;
  }

  /**
   * Creates a soft reference instead of a hard one linking to the body of the
   * contained message. The message must have been saved previously.
   */
  public void releaseFullMessage() {
    if (soft) {
      bodySoftRef = new SoftReference(getMsg().body);
      msg.body = null;
    }
  }

  /** Returns the message type. */
  public int getType() {
    return getMsg().type;
  }

  /** Returns the message identifier. */
  public String getId() {
    return getMsg().id;
  }

  /** Sets the message identifier. */ 
  public void setIdentifier(String id) {
    getMsg().id = id;
  }

  /** Returns <code>true</code> if the message is persistent. */
  public boolean isPersistent() {
    return getMsg().persistent;
  }

  /** Sets the message persistence mode. */
  public void setPersistent(boolean persistent) {
    getMsg().persistent = persistent;
  }

  /** Returns the message priority. */
  public int getPriority() {
    return getMsg().priority;
  }

  /**
   * Sets the message priority.
   *
   * @param priority  Priority value: 0 the lowest, 9 the highest, 4 normal.
   */ 
  public void setPriority(int priority) {
    if (priority >= 0 && priority <= 9)
      getMsg().priority = priority;
  }

  /** Returns the message expiration time. */
  public long getExpiration() {
    return getMsg().expiration;
  }

  /**
   * Sets the message expiration.
   *
   * @param expiration	The expiration time.
   */
  public void setExpiration(long expiration) {
    if (expiration >= 0)
      getMsg().expiration = expiration;
  }

  /** Returns the message time stamp. */
  public long getTimestamp() {
    return getMsg().timestamp;
  }

  /** Sets the message time stamp. */
  public void setTimestamp(long timestamp) {
    getMsg().timestamp = timestamp;
  }

  /** Returns the message correlation identifier. */
  public final String getCorrelationId() {
    return getMsg().correlationId;
  }

  /** Sets the message correlation identifier. */
  public void setCorrelationId(String correlationId) {
    getMsg().correlationId = correlationId;
  }

  /** Returns the message delivery count.*/
  public int getDeliveryCount() {
    return getMsg().deliveryCount;
  }

  /** Sets the message delivery count. */
  public void setDeliveryCount(int deliveryCount) {
    getMsg().deliveryCount = deliveryCount;
  }

  /** Increments the message delivery count. */
  public void incDeliveryCount() {
    getMsg().deliveryCount += 1;
  }

  /** Sets the message redelivered flag. */
  public void setRedelivered() {
    getMsg().redelivered = true;
  }
  
  /** Sets the message delivery time */
  public void setDeliveryTime(long deliveryTime) {
    getMsg().deliveryTime = deliveryTime;
  }
  
  /** Returns the message delivery time*/
  public long getDeliveryTime() {
    return getMsg().deliveryTime;
  }
  
  public /*@ pure @*/ long getOrder() {
    return order;
  }

   /*@ modifies this.order; @*/
  public void setOrder(long order) {
    this.order = order;
  }

  public /*@ pure @*/ int getAcksCounter() {
    return acksCounter;
  }

  public void incAcksCounter() {
    this.acksCounter++;
  }

  public /*@ pure @*/ int getDurableAcksCounter() {
    return durableAcksCounter;
  }

  public void incDurableAcksCounter() {
    this.durableAcksCounter++;
  }

  /**
   * Get the clientID
   * @return the clientID
   */
  public String getClientID() {
    return getMsg().clientID;
  }

  /**
   * Sets a property value.
   * If the value is not a Java primitive object its string representation is used.
   *
   * @param name  The property name.
   * @param value  The property value.
   */
  public void setObjectProperty(String name, Object value) {
    getMsg().setProperty(name, value);
  }

  //   /**
  //    * Sets an object as the body of the message. 
  //    * AF: Used to wrap administration message !!
  //    *
  //    * @exception IOException  In case of an error while setting the object.
  //    */
  //   public void setObject(Object object) throws IOException {
  //     msg.type = Message.OBJECT;

  //     if (object == null) {
  //       msg.body = null;
  //     } else {
  //       ByteArrayOutputStream baos = new ByteArrayOutputStream();
  //       ObjectOutputStream oos = new ObjectOutputStream(baos);
  //       oos.writeObject(object);
  //       oos.flush();
  //       msg.body = baos.toByteArray();
  //       oos.close();
  //       baos.close();
  //     }
  //   }

  //   /**
  //    * Returns the object body of the message.
  //    * AF: Used to wrap administration message !!
  //    *
  //    * @exception IOException  In case of an error while getting the object.
  //    * @exception ClassNotFoundException  If the object class is unknown.
  //    */
  //   public Object getObject() throws Exception {
  //     // AF: May be, we should verify that it is an Object message!!
  //     if (msg.body == null) return null;

  //     ByteArrayInputStream bais = new ByteArrayInputStream(msg.body);
  //     ObjectInputStream ois = new ObjectInputStream(bais);
  //     Object obj = ois.readObject();
  //     ois.close();

  //     return obj;
  //   }

  /**
   * Returns <code>true</code> if the message is valid.
   * The message is valid if not expired.
   *
   * @param currentTime	The current time to verify the expiration time.
   */
  public boolean isValid(long currentTime) {
    return (getMsg().expiration <= 0) || (getMsg().expiration > currentTime);
  }

  /**
   * Return true if the message has an expiration delay.
   * 
   * @return true if the message has an expiration delay.
   */
  public final boolean hasExpiration() {
    return getMsg().expiration > 0;
  }
  
  /** Name used to store the message */
  transient String txname = null;

  public void setTxName(String txname) {
    this.txname = txname;
  }

  public String getTxName() {
    return txname;
  }

  public static Message load(String txname) throws IOException, ClassNotFoundException {
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "Message.load:" + txname);
    Message m = (Message) AgentServer.getTransaction().load(txname);
    m.txname = txname;
    return m;
  }

  /**
   * Method used to save the initial state of the message.
   */
  public void save() {
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "Message.save:" + txname);

    if (!isPersistent())
      // Loads the message if needed
      return;
    
    if (soft) {
      byte[] body = msg.body;
      // sets the body to null to save it in an other file
      msg.body = null;
      try {
        AgentServer.getTransaction().create(this, txname);
      } catch (IOException exc) {
        logger.log(BasicLevel.ERROR, "Message named [" + txname + "] could not be saved", exc);
      }
      // save the body
      try {
        // The body is RO do not copy it.
        AgentServer.getTransaction().saveByteArray(body, null, txname + "B", false, true);
      } catch (IOException exc) {
        logger.log(BasicLevel.ERROR, "Message named [" + txname + "] could not be saved", exc);
      }
      msg.body = body;
    } else {
      try {
        AgentServer.getTransaction().create(this, txname);
      } catch (IOException exc) {
        logger.log(BasicLevel.ERROR, "Message named [" + txname + "] could not be saved", exc);
      }
    }
  }

  /**
   * Method used to save the header of a message after modification.
   * The body of a message should never be saved.
   */
  public void saveHeader() {
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "Message.saveHeader:" + txname);

    if (!isPersistent())
      // Loads the message if needed
      return;
    
    if (soft) {
      byte[] body = msg.body;
      // sets the body to null to not save it
      msg.body = null;
      try {
        AgentServer.getTransaction().save(this, txname);
      } catch (IOException exc) {
        logger.log(BasicLevel.ERROR, "Message named [" + txname + "] could not be saved", exc);
      }
      msg.body = body;
    } else {
      try {
        AgentServer.getTransaction().save(this, txname);
      } catch (IOException exc) {
        logger.log(BasicLevel.ERROR, "Message named [" + txname + "] could not be saved", exc);
      }
    }
  }

  public void delete() {
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "Message.delete:" + txname);

    if (!isPersistent()) return;
    
    AgentServer.getTransaction().delete(txname);
    if (soft) {
      AgentServer.getTransaction().delete(txname + "B");
    }
  }

  /**
   * Loads all persisted messages.
   * The returned list is sorted according to the message order and priority if any.
   * 
   * If there is more message to restore than max, the Message structure is created but
   * the message is not loaded 
   * 
   * @param msgTxname prefix of messages to load.
   * @param max       the maximum number of messages to load
   * @return          a vector containing the loaded messages.
   */
  public static Vector<Message> loadAll(String msgTxname, int max) {
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "Message.loadAll() " + msgTxname);

    Vector<Message> messages = new Vector<Message>();
    Transaction tx = AgentServer.getTransaction();

    // TODO (AF): Use DBTransaction optimized loadAll method to get all messages from a unique request.
    // Currently this optimization results in an issue if messages swap is allowed, it needs to filter
    // body objects that have the same prefix).
    if (tx.useLoadAll() && useLoadALL) {
      // Retrieving the saved messages from persistency. 
      long start = System.currentTimeMillis();
      Map<String, Message> msgsMap = new HashMap<String, Message>(32000);
      tx.loadAll(msgTxname + "", msgsMap);
      long end = System.currentTimeMillis();

      logger.log(BasicLevel.INFO,
                 "Message.loadAll: all messages loaded (" + msgsMap.size() + ") -> " + (end - start));

      start = System.currentTimeMillis();
      for (Map.Entry<String, Message> entry : msgsMap.entrySet()) {
        String name = entry.getKey();
        if (name.charAt(name.length() - 1) == 'B')
          // Never happen, the corresponding objects are bytes arrays and throw exceptions in loadAll
          continue;
        Message msg = entry.getValue();
        msg.txname = name;
        if (msg.msg == null) {
          // TODO (AF): retrieve body in Map && msg.soft ?
        }
        messages.add(msg);
      }
      end = System.currentTimeMillis();
      logger.log(BasicLevel.INFO,
                 "Message.loadAll: insert (" + messages.size() + ") -> " + (end - start));

      start = System.currentTimeMillis();
      // Sort messages list in reverse order to optimize queue restoration (JORAM-358)
      Collections.sort(messages);
      end = System.currentTimeMillis();
      logger.log(BasicLevel.INFO,
                 "Message.loadAll: sort -> " + (end - start));
    } else {
      // Retrieving the names of the persistence message previously saved. 
      long start = System.currentTimeMillis();
      String[] names = tx.getList(msgTxname);
      long end = System.currentTimeMillis();
      logger.log(BasicLevel.INFO,
                 "Message.loadAll: retrieves messages list (" + names.length + ") -> " + (end - start));
      logger.log(BasicLevel.DEBUG,
                 "Message.loadAll: names = " + Arrays.toString(names));

      // Retrieving the messages individually persisted.
      int nb = 0;
      Message msg;
      int priority = -1;
      boolean samePriorities = true;
      start = System.currentTimeMillis();
      for (String name : names) {
        if (name.charAt(name.length() - 1) == 'B')
          continue;

        try {
          if (nb < max) {
            msg = (Message) tx.load(name);
            if (msg.msg.priority != priority) {
              if (priority == -1) {
                // 1st message
                priority = msg.msg.priority;
              } else {
                samePriorities = false;
              }
            }
            if (logger.isLoggable(BasicLevel.DEBUG)) {
              // Verify that the order coded in txname is the right one.
              int idx = name.lastIndexOf('_');
              long order = Long.parseLong(name.substring(idx+1));
              if (order != msg.order)
                logger.log(BasicLevel.WARN,
                           "Message.loadAll: Message " + name + " -> " + msg.order + " != " + order);
            }
          } else {
            msg = new Message();
            // Get order from txname
            int idx = name.lastIndexOf('_');
            msg.order = Long.parseLong(name.substring(idx+1));
          }
          msg.txname = name;
          nb += 1;

          if (logger.isLoggable(BasicLevel.DEBUG))
            logger.log(BasicLevel.DEBUG,
                       "Message.loadAll: names[" + nb + "] = " + msg.txname);
          messages.add(msg);
        } catch (Exception exc) {
          logger.log(BasicLevel.ERROR,
                     "Message.loadAll: Message named [" + name + "] can not be loaded", exc);
        }
      }
      end = System.currentTimeMillis();
      logger.log(BasicLevel.INFO,
                 "Message.loadAll: all messages loaded -> " + (end - start));

      start = System.currentTimeMillis();
      // Sort messages list in reverse order to optimize queue restoration (JORAM-358)
      if (samePriorities) {
        Collections.sort(messages, new Comparator<Message>() {
          @Override
          public int compare(Message msg1, Message msg2) {
            return Long.compare(msg1.order, msg2.order);
          }
        });
      } else {
        Collections.sort(messages, new Comparator<Message>() {
          @Override
          public int compare(Message msg1, Message msg2) {
            if ((msg1.msg != null) && (msg2.msg != null)) {
              return Long.compare(msg1.order, msg2.order);
            }
            if ((msg1.msg != null) && (msg2.msg == null)) {
              return -1;
            }
            if ((msg1.msg == null) && (msg2.msg != null)) {
              return 1;
            }
            return Long.compare(msg1.order, msg2.order);
          }
        });
      }
      end = System.currentTimeMillis();
      logger.log(BasicLevel.INFO,
                 "Message.loadAll: sort -> " + (end - start));
    }
    
    return messages;
  }

  /** Deletes all persisted objects. */
  public static void deleteAll(String msgTxname) {
    if (logger.isLoggable(BasicLevel.DEBUG))
      logger.log(BasicLevel.DEBUG, "MessagePersistenceModule.deleteAll() " + msgTxname);

    Transaction tx = AgentServer.getTransaction();

    // Retrieving the names of the persistence message previously saved. 
    String[] names = tx.getList(msgTxname);

    // Deleting the message.
    for (int i = 0; i < names.length; i++) {
      tx.delete(names[i]);
      tx.delete(names[i] + "B");
    }
  }

  /* ***** ***** ***** ***** ***** ***** ***** *****
   * Serializable interface
   * ***** ***** ***** ***** ***** ***** ***** ***** */
  private void writeObject(ObjectOutputStream out) throws IOException {
    out.writeLong(order);
    out.writeBoolean(soft);

    msg.writeHeaderTo(out);
    if (msg.bodyLength < 0) {
      StreamUtil.writeTo(msg.body, out);
    } else {
      StreamUtil.writeTo(msg.body, msg.bodyOffset, msg.bodyLength, out);
    }
  }

  /**
   * @throws ClassNotFoundException  
   */
   /*@ modifies this.durableAcksCounter; @*/
   /*@ modifies this.acksCounter; @*/
   /*@ modifies this.order; @*/
  private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
    order = in.readLong();
    soft = in.readBoolean();

    acksCounter = 0;
    durableAcksCounter = 0;

    msg = new org.objectweb.joram.shared.messages.Message();
    msg.readHeaderFrom(in);
    msg.body = StreamUtil.readByteArrayFrom(in);
    msg.bodyOffset = 0;
    if (msg.body != null) {
      msg.bodyLength = msg.body.length;
    }
  }

  public String getText() {
    try {
	    return getMsg().getText();
    } catch (Exception e) {
    	if (logger.isLoggable(BasicLevel.INFO))
        logger.log(BasicLevel.INFO, "getText()", e);
	    return null;
    }
  }

  public boolean isRedelivered() {
    return getMsg().redelivered;
  }

  public Map getProperties() {
    if (getMsg().properties == null)
      return null;

    Map props = new HashMap();
    Enumeration enu = msg.properties.keys();
    while (enu.hasMoreElements()) {
      String key = (String) enu.nextElement();
      props.put(key, msg.properties.get(key).toString());
    }
    return props;
  }
  
  public int getEncodableClassId() {
    return JoramHelper.MESSAGE_CLASS_ID;
  }
  
  public int getEncodedSize() throws Exception {
    int encodedSize = 0;
    encodedSize += LONG_ENCODED_SIZE + BOOLEAN_ENCODED_SIZE;
    encodedSize += msg.getEncodedSize();
    return encodedSize;
  }

  public void encode(Encoder encoder) throws Exception {
    encoder.encodeUnsignedLong(order);
    encoder.encodeBoolean(soft);
    msg.encode(encoder);
  }

   /*@ modifies this.order; @*/
   /*@ modifies this.durableAcksCounter; @*/
   /*@ modifies this.acksCounter; @*/
  public void decode(Decoder decoder) throws Exception {
    order = decoder.decodeUnsignedLong();
    soft = decoder.decodeBoolean();
    acksCounter = 0;
    durableAcksCounter = 0;
    msg = new org.objectweb.joram.shared.messages.Message();
    msg.decode(decoder);
  }
  
  public static class MessageFactory implements EncodableFactory {
    public Encodable createEncodable() {
      return new Message();
    }
  }
}
